"""This module provides helper functions for smarter ESI requests with django-esi
    
    Added features for all ESI requests:
    - Automatic page retry on 502, 503, 504 up to max retries with exponential backoff
    - Automatic retrieval of all pages
    - Automatic retrieval of variants for all requested languages
"""

import logging
from time import sleep

from bravado.exception import HTTPBadGateway
from bravado.exception import HTTPServiceUnavailable
from bravado.exception import HTTPGatewayTimeout

from django.conf import settings

from esi.clients import esi_client_factory
from esi.models import Token

from .. import __title__
from ..app_settings import SR_ESI_TIMEOUT_ENABLED
from ..utils import LoggerAddTag


logger = LoggerAddTag(logging.getLogger(__name__), __title__)

ESI_MAX_RETRIES = 3
ESI_RETRY_SLEEP_SECS = 1

_my_esi_client = None


def esi_fetch(
    esi_path: str,
    args: dict = None,
    has_pages: bool = False,
    token: Token = None,
    esi_client: object = None,
    logger_tag: str = None,
) -> dict:
    """returns an response object from ESI, will retry on some HTTP errors.
    will automatically return all pages if requested

    Args:
    - esi_path: Full path of esi route, 
    e.g. ``Universe.get_universe_categories_category_id``
    - args: arguments for ESI method as dict, e.g. ``{'category_id': 65}``
    - has_pages: When set to True will assume endpoint supports paging
    - token: esi token from django-esi to be used with request
    - esi_client: esi client object from django-esi to be used for request 
    instead of default esi client from this module
    - logger_tag: every log message will start with this text in brackets
    """
    _, request_object = _fetch_main(
        esi_path=esi_path,
        args=args,
        languages=None,
        has_pages=has_pages,
        esi_client=esi_client,
        token=token,
        logger_tag=logger_tag,
    ).popitem()
    return request_object


def esi_fetch_with_localization(
    esi_path: str,
    languages: set,
    args: dict = None,
    has_pages: bool = False,
    esi_client: object = None,
    token: Token = None,
    logger_tag: str = None,
) -> dict:
    """returns dict of response objects from ESI
    will contain one full object items for each language if supported or just one
    will retry on bad request responses from ESI
    will automatically return all pages if requested

    Args:
    - esi_path: Full path of esi route, 
    e.g. ``Universe.get_universe_categories_category_id``
    - languages: languages to be retrieved from ESI as codes, 
    should match official codes supported by ESI, e.g. ``{'de', 'ko'}``
    - args: arguments for ESI method as dict, e.g. ``{'category_id': 65}``
    - has_pages: When set to True will assume endpoint supports paging
    - token: esi token from django-esi to be used with request
    - esi_client: esi client object from django-esi to be used for request 
    instead of default esi client from this module
    - logger_tag: every log message will start with this text in brackets
    """
    return _fetch_main(
        esi_path=esi_path,
        args=args,
        languages=languages,
        has_pages=has_pages,
        esi_client=esi_client,
        token=token,
        logger_tag=logger_tag,
    )


def _fetch_main(
    esi_path: str,
    args: dict,
    languages: set,
    has_pages: bool,
    esi_client: object,
    token: Token,
    logger_tag: str,
) -> dict:
    """returns dict of response objects from ESI with localization"""

    if not args:
        args = {}

    if not languages:
        has_localization = False
        languages = {"dummy"}
    else:
        has_localization = True

    response_objects = dict()
    for language in languages:
        if has_localization:
            args["language"] = language
        response_objects[language] = _fetch_with_paging(
            esi_path=esi_path,
            args=args,
            has_pages=has_pages,
            esi_client=esi_client,
            token=token,
            logger_tag=logger_tag,
        )

    return response_objects


def _fetch_with_paging(
    esi_path: str,
    args: dict,
    has_pages: bool = False,
    esi_client: object = None,
    token: Token = None,
    logger_tag: str = None,
) -> dict:
    """fetches esi objects incl. all pages if requested and returns them"""
    response_object, pages = _fetch_with_retries(
        esi_path=esi_path,
        args=args,
        has_pages=has_pages,
        esi_client=esi_client,
        token=token,
        logger_tag=logger_tag,
    )
    if has_pages:
        for page in range(2, pages + 1):
            response_object_page, _ = _fetch_with_retries(
                esi_path=esi_path,
                args=args,
                has_pages=has_pages,
                page=page,
                pages=pages,
                esi_client=esi_client,
                token=token,
                logger_tag=logger_tag,
            )
            response_object += response_object_page

    return response_object


def _esi_client() -> object:
    """returns the singular esi client used in this module"""
    global _my_esi_client

    if not _my_esi_client:
        logger.info("Initializing esi client for esi_fetch....")
        _my_esi_client = esi_client_factory()

    return _my_esi_client


def _fetch_with_retries(
    esi_path: str,
    args: dict,
    has_pages: bool = False,
    page: int = None,
    pages: int = None,
    esi_client: object = None,
    token: Token = None,
    logger_tag: str = None,
) -> tuple:
    """Returns response object and pages from ESI, retries on 502s"""

    esi_category, esi_method_name, log_message_base = _prepare_esi_request(
        esi_path=esi_path,
        args=args,
        has_pages=has_pages,
        page=page,
        pages=pages,
        esi_client=esi_client,
        token=token,
    )
    response_object, pages = _execute_esi_request(
        esi_category=esi_category,
        esi_method_name=esi_method_name,
        args=args,
        has_pages=has_pages,
        logger_tag=logger_tag,
        log_message_base=log_message_base,
    )
    return response_object, pages


def _prepare_esi_request(
    esi_path: str,
    args: dict,
    has_pages: bool = False,
    page: int = None,
    pages: int = None,
    esi_client: object = None,
    token: Token = None,
):
    """parses and validates input for esi request"""
    esi_path_parts = esi_path.split(".")
    if len(esi_path_parts) != 2:
        raise ValueError("Invalid esi_path")
    esi_category_name = esi_path_parts[0]
    esi_method_name = esi_path_parts[1]
    if not esi_client:
        esi_client = _esi_client()
    if not hasattr(esi_client, esi_category_name):
        raise ValueError("Invalid ESI category: %s" % esi_category_name)
    esi_category = getattr(esi_client, esi_category_name)
    if not hasattr(esi_category, esi_method_name):
        raise ValueError(
            "Invalid ESI method for %s category: %s"
            % (esi_category_name, esi_method_name)
        )
    log_message_base = "Fetching from ESI: {}".format(esi_path)
    if settings.DEBUG:
        log_message_base += "({})".format(
            ", ".join([str(k) + "=" + str(v) for k, v in args.items()])
        )
    if has_pages:
        if not page:
            page = 1
        args["page"] = page
        log_message_base += " - Page {}/{}".format(page, pages if pages else "?")
    if token:
        if token.expired:
            token.refresh()
        args["token"] = token.access_token

    return esi_category, esi_method_name, log_message_base


def _execute_esi_request(
    esi_category: str,
    esi_method_name: str,
    args: dict,
    has_pages: bool,
    logger_tag: str,
    log_message_base: str,
):
    """make request to ESI
    
    returns request object and total number of pages to retrieve
    """
    add_prefix = _make_logger_prefix(logger_tag)
    logger.info(add_prefix(log_message_base))
    response_object = dict()
    pages = 0
    for retry_count in range(ESI_MAX_RETRIES + 1):
        if retry_count > 0:
            logger.warn(
                add_prefix(
                    "{} - Retry {} / {}".format(
                        log_message_base, retry_count, ESI_MAX_RETRIES
                    )
                )
            )
        try:
            operation = getattr(esi_category, esi_method_name)(**args)
            result_args = {"timeout": (5, 30)} if SR_ESI_TIMEOUT_ENABLED else {}
            if has_pages:
                if hasattr(operation, "request_config"):
                    operation.request_config.also_return_response = True
                    response_object, response = operation.result(**result_args)
                elif hasattr(operation, "also_return_response"):
                    operation.also_return_response = True
                    response_object, response = operation.result(**result_args)
                else:
                    logger.warning(
                        "django-esi API is not fully compatible. "
                        "Falling back to fetching first page only from ESI"
                    )
                    response_object = operation.result(**result_args)
                    response = None

                if response and "x-pages" in response.headers:
                    pages = int(response.headers["x-pages"])
                else:
                    pages = 0
            else:
                response_object = operation.result(**result_args)
                pages = 0
            break

        except (HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable) as ex:
            logger.warn(
                add_prefix(
                    "HTTP error while trying to "
                    "fetch response_object from ESI: {}".format(ex)
                )
            )
            if retry_count < ESI_MAX_RETRIES:
                sleep_seconds = (ESI_RETRY_SLEEP_SECS * retry_count) ** 2
                logger.info(
                    add_prefix(
                        "Waiting {} seconds until next retry".format(sleep_seconds)
                    )
                )
                sleep(sleep_seconds)
            else:
                raise ex

    return response_object, pages


def _make_logger_prefix(tag: str = None):
    """creates a function to add logger prefix"""
    return lambda text: "{}{}".format((tag + ": ") if tag else "", text)



from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger

from django.contrib.auth.decorators import login_required, permission_required
from django.http import Http404, HttpResponse, JsonResponse
from django.shortcuts import redirect, render
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from esi.decorators import token_required

from . import __title__
from .app_settings import (
    SR_CORPORATIONS_ENABLED,
    SR_OPERATION_MODE,
    STANDINGS_API_CHARID,
    SR_NOTIFICATIONS_ENABLED,
)
from .decorators import token_required_by_state
from .helpers.evecharacter import EveCharacterHelper
from .helpers.evecorporation import EveCorporation
from .helpers.eveentity import EveEntityHelper
from .helpers.viewcache import cache_view_pilots_json, cache_view_groups_json
from .helpers.writers import UnicodeWriter
from .models import (
    CharacterContact,
    CorporationContact,
    ContactSet,
    EveEntity,
    StandingRequest,
    StandingRevocation,
    models,
)
from .tasks import update_all
from .utils import LoggerAddTag, messages_plus

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

DEFAULT_ICON_SIZE = 32
CACHED_PAGES_MINUTES = 10


class HttpResponseNoContent(HttpResponse):
    """Special HTTP response with no content, just headers.

    The content operations are ignored.
    """

    def __init__(self, content="", mimetype=None, status=None, content_type=None):
        super().__init__(status=204)

        # although we don't define a content-type, base class sets a
        # default one -- remove it, we're not returning content
        if "content-type" in self._headers:
            del self._headers["content-type"]

    def _set_content(self, value):
        pass

    def _get_content(self, value):
        pass


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def index_view(request):
    logger.debug("Start index_view request")
    context = {
        "app_title": __title__,
        "operation_mode": SR_OPERATION_MODE,
        "corporations_enabled": SR_CORPORATIONS_ENABLED,
    }
    return render(request, "standingsrequests/index.html", context)


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def create_request_debug(request):
    logger.debug("Start index_view request")
    context = {
        "app_title": __title__,
        "operation_mode": SR_OPERATION_MODE,
        "corporations_enabled": SR_CORPORATIONS_ENABLED,
    }
    return render(request, "standingsrequests/create_request_debug.html", context)


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def partial_request_entities(request):
    logger.debug("Start partial_request_entities request")
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        return render(
            request,
            "standingsrequests/error.html",
            {"app_title": __title__, "operation_mode": SR_OPERATION_MODE,},
        )

    eve_characters_qs = EveEntityHelper.get_characters_by_user(request.user)
    character_contacts_qs = contact_set.charactercontact_set.filter(
        contact_id__in=eve_characters_qs.values_list("character_id", flat=True)
    )
    characters_data = list()
    for character in eve_characters_qs:
        try:
            standing = character_contacts_qs.get(
                contact_id=character.character_id
            ).standing
        except CharacterContact.DoesNotExist:
            standing = None

        characters_data.append(
            {
                "character": character,
                "standing": standing,
                "pendingRequest": StandingRequest.objects.has_pending_request(
                    character.character_id
                ),
                "pendingRevocation": StandingRevocation.objects.has_pending_request(
                    character.character_id
                ),
                "requestActioned": StandingRequest.objects.has_actioned_request(
                    character.character_id
                ),
                "inOrganisation": ContactSet.is_character_in_organisation(character),
                "hasRequiredScopes": StandingRequest.has_required_scopes_for_request(
                    character, user=request.user, quick_check=True
                ),
                "hasStanding": StandingRequest.objects.filter(
                    contact_id=character.character_id,
                    user=request.user,
                    is_effective=True,
                ).exists(),
            }
        )

    corporations_data = list()
    if SR_CORPORATIONS_ENABLED:
        corporation_ids = set(
            eve_characters_qs.exclude(
                corporation_id__in=ContactSet.corporation_ids_in_organization()
            )
            .exclude(alliance_id__in=ContactSet.alliance_ids_in_organization())
            .values_list("corporation_id", flat=True)
        )
        corporation_contacts_qs = contact_set.corporationcontact_set.filter(
            contact_id__in=corporation_ids
        )
        for corporation in EveCorporation.get_many_by_id(corporation_ids):
            if corporation and not corporation.is_npc:
                try:
                    standing = corporation_contacts_qs.get(
                        contact_id=corporation.corporation_id
                    ).standing
                except CorporationContact.DoesNotExist:
                    standing = None

                corporations_data.append(
                    {
                        "token_count": corporation.member_tokens_count_for_user(
                            request.user, quick_check=True
                        ),
                        "corp": corporation,
                        "standing": standing,
                        "pendingRequest": StandingRequest.objects.has_pending_request(
                            corporation.corporation_id
                        ),
                        "pendingRevocation": StandingRevocation.objects.has_pending_request(
                            corporation.corporation_id
                        ),
                        "requestActioned": StandingRequest.objects.has_actioned_request(
                            corporation.corporation_id
                        ),
                        "hasStanding": StandingRequest.objects.filter(
                            contact_id=corporation.corporation_id, is_effective=True,
                        ).exists(),
                    }
                )

        corporations_data.sort(key=lambda x: x["corp"].corporation_name)

    organization = ContactSet.standings_source_entity()
    render_items = {
        "app_title": __title__,
        "characters": characters_data,
        "corps": corporations_data,
        "operation_mode": SR_OPERATION_MODE,
        "corporations_enabled": SR_CORPORATIONS_ENABLED,
        "organization": organization,
        "organization_image_url": organization.icon_url(size=DEFAULT_ICON_SIZE),
        "authinfo": {"main_char_id": request.user.profile.main_character.character_id},
    }
    return render(
        request, "standingsrequests/partials/_request_entities.html", render_items
    )


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def request_pilot_standing(request, character_id):
    """For a user to request standings for their own pilots"""
    ok = True
    character_id = int(character_id)
    logger.debug(
        "Standings request from user %s for characterID %d", request.user, character_id
    )
    character = EveCharacter.objects.get_character_by_id(character_id)
    if not character or not EveEntityHelper.is_character_owned_by_user(
        character_id, request.user
    ):
        logger.warning(
            "User %s does not own Pilot ID %d, forbidden", request.user, character_id
        )
        ok = False
    elif StandingRequest.objects.has_pending_request(
        character_id
    ) or StandingRevocation.objects.has_pending_request(character_id):
        logger.warning("Contact ID %d already has a pending request", character_id)
        ok = False
    elif not StandingRequest.has_required_scopes_for_request(
        character=character, user=request.user
    ):
        ok = False
    else:
        sr = StandingRequest.objects.add_request(
            user=request.user,
            contact_id=character_id,
            contact_type=StandingRequest.CHARACTER_CONTACT_TYPE,
        )
        try:
            contact_set = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.warning("Failed to get a contact set")
            ok = False
        else:
            if contact_set.character_has_satisfied_standing(character_id):
                sr.mark_actioned(user=None)
                sr.mark_effective()

    if not ok:
        messages_plus.warning(
            request,
            "An unexpected error occurred when trying to process "
            "your standing request for %s. Please try again."
            % EveEntity.objects.get_name(character_id),
        )

    return redirect("standingsrequests:index")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def remove_pilot_standing(request, character_id):
    """
    Handles both removing requests and removing existing standings
    """
    character_id = int(character_id)
    ok = True
    logger.debug(
        "remove_pilot_standing called by %s for character %d",
        request.user,
        character_id,
    )
    character = EveCharacter.objects.get_character_by_id(character_id)
    if not character or not EveEntityHelper.is_character_owned_by_user(
        character_id, request.user
    ):
        logger.warning(
            "User %s does not own Pilot ID %d, forbidden", request.user, character_id
        )
        ok = False
    elif ContactSet.is_character_in_organisation(character):
        logger.warning(
            "Character %d of user %s is in organization. Can not remove standing",
            character_id,
            request.user,
        )
        ok = False
    elif not StandingRevocation.objects.has_pending_request(character_id):
        if StandingRequest.objects.has_pending_request(
            character_id
        ) or StandingRequest.objects.has_actioned_request(character_id):
            logger.debug(
                "Removing standings requests for character ID %d by user %s",
                character_id,
                request.user,
            )
            StandingRequest.objects.remove_requests(character_id)
        else:
            try:
                contact_set = ContactSet.objects.latest()
            except ContactSet.DoesNotExist:
                logger.warning("Failed to get a contact set")
                ok = False
            else:
                if contact_set.character_has_satisfied_standing(character_id):
                    logger.debug(
                        "Creating standings revocation for character ID %d by user %s",
                        character_id,
                        request.user,
                    )
                    StandingRevocation.objects.add_revocation(
                        contact_id=character_id,
                        contact_type=StandingRevocation.CHARACTER_CONTACT_TYPE,
                        user=request.user,
                    )
                else:
                    logger.debug("No standings exist for characterID %d", character_id)
    else:
        logger.debug(
            "User %s already has a pending standing revocation for character %d",
            request.user,
            character_id,
        )

    if not ok:
        messages_plus.warning(
            request,
            "An unexpected error occurred when trying to process "
            "your request to revoke standing for %s. Please try again."
            % EveEntity.objects.get_name(character_id),
        )

    return redirect("standingsrequests:index")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def request_corp_standing(request, corporation_id):
    """
    For a user to request standings for their own corp
    """
    corporation_id = int(corporation_id)
    ok = True
    logger.debug(
        "Standings request from user %s for corpID %d", request.user, corporation_id
    )
    if StandingRequest.can_request_corporation_standing(corporation_id, request.user):
        if not StandingRequest.objects.has_pending_request(
            corporation_id
        ) and not StandingRevocation.objects.has_pending_request(corporation_id):
            StandingRequest.objects.add_request(
                user=request.user,
                contact_id=corporation_id,
                contact_type=StandingRequest.CORPORATION_CONTACT_TYPE,
            )
        else:
            # Pending request, not allowed
            logger.warning(
                "Contact ID %d already has a pending request", corporation_id
            )
            ok = False
    else:
        logger.warning(
            "User %s does not have enough keys for corpID %d, forbidden",
            request.user,
            corporation_id,
        )
        ok = False

    if not ok:
        messages_plus.warning(
            request,
            "An unexpected error occurred when trying to process "
            "your standing request for %s. Please try again."
            % EveEntity.objects.get_name(corporation_id),
        )

    return redirect("standingsrequests:index")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
def remove_corp_standing(request, corporation_id):
    """
    Handles both removing corp requests and removing existing standings
    """
    corporation_id = int(corporation_id)
    ok = True
    logger.debug("remove_corp_standing called by %s", request.user)
    try:
        st_req = StandingRequest.objects.get(contact_id=corporation_id)
    except StandingRequest.DoesNotExist:
        ok = False
    else:
        if st_req.user == request.user:
            if (
                StandingRequest.objects.has_pending_request(corporation_id)
                or StandingRequest.objects.has_actioned_request(corporation_id)
            ) and not StandingRevocation.objects.has_pending_request(corporation_id):
                logger.debug(
                    "Removing standings requests for corpID %d by user %s",
                    corporation_id,
                    request.user,
                )
                StandingRequest.objects.remove_requests(corporation_id)
            else:
                try:
                    contact_set = ContactSet.objects.latest()
                except ContactSet.DoesNotExist:
                    logger.warning("Failed to get a contact set")
                    ok = False
                else:
                    if contact_set.corporation_has_satisfied_standing(corporation_id):
                        # Manual revocation required
                        logger.debug(
                            "Creating standings revocation for corpID %d by user %s",
                            corporation_id,
                            request.user,
                        )
                        StandingRevocation.objects.add_revocation(
                            contact_id=corporation_id,
                            contact_type=StandingRevocation.CORPORATION_CONTACT_TYPE,
                            user=request.user,
                        )
                    else:
                        logger.debug(
                            "Can remove standing - no standings exist for corpID %d",
                            corporation_id,
                        )
                        ok = False

        else:
            logger.warning(
                "User %s tried to remove standings for corpID %d but was not permitted",
                request.user,
                corporation_id,
            )
            ok = False

    if not ok:
        messages_plus.warning(
            request,
            "An unexpected error occurred when trying to process "
            "your request to revoke standing for %s. Please try again."
            % EveEntity.objects.get_name(corporation_id),
        )

    return redirect("standingsrequests:index")


###########################
# Views character and groups #
###########################
@login_required
@permission_required("standingsrequests.view")
def view_pilots_standings(request):
    logger.debug("view_pilot_standings called by %s", request.user)
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contact_set = None
    finally:
        organization = ContactSet.standings_source_entity()
        last_update = contact_set.date if contact_set else None
        pilots_count = contact_set.charactercontact_set.count() if contact_set else None

    return render(
        request,
        "standingsrequests/view_pilots.html",
        {
            "lastUpdate": last_update,
            "app_title": __title__,
            "operation_mode": SR_OPERATION_MODE,
            "organization": organization,
            "pilots_count": pilots_count,
        },
    )


@login_required
@permission_required("standingsrequests.view")
def view_pilots_standings_json(request):
    logger.debug("view_pilot_standings_json called by %s", request.user)

    def generate_characters_data():
        try:
            contacts = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            contacts = ContactSet()

        characters_data = list()
        character_contacts = contacts.charactercontact_set.all().order_by("name")
        for p in character_contacts:
            character = EveCharacter.objects.get_character_by_id(p.contact_id)
            try:
                user = character.character_ownership.user
            except AttributeError:
                character = EveCharacterHelper(p.contact_id)
                main = None
                state = ""
                has_required_scopes = None
            else:
                user = character.character_ownership.user
                main = user.profile.main_character
                state = user.profile.state.name if user.profile.state else ""
                has_required_scopes = StandingRequest.has_required_scopes_for_request(
                    character=character, user=user, quick_check=True
                )
            finally:
                character_icon_url = character.portrait_url(DEFAULT_ICON_SIZE)
                corporation_id = character.corporation_id if character else None
                corporation_name = character.corporation_name if character else None
                corporation_ticker = character.corporation_ticker if character else None
                alliance_id = character.alliance_id if character else None
                alliance_name = character.alliance_name if character else None
                main_character_name = main.character_name if main else None
                main_character_ticker = main.corporation_ticker if main else None
                main_character_icon_url = (
                    main.portrait_url(DEFAULT_ICON_SIZE) if main else None
                )
                labels = [label.name for label in p.labels.all()]

            characters_data.append(
                {
                    "character_id": p.contact_id,
                    "character_name": p.name,
                    "character_icon_url": character_icon_url,
                    "corporation_id": corporation_id,
                    "corporation_name": corporation_name,
                    "corporation_ticker": corporation_ticker,
                    "alliance_id": alliance_id,
                    "alliance_name": alliance_name,
                    "has_required_scopes": has_required_scopes,
                    "state": state,
                    "main_character_name": main_character_name,
                    "main_character_ticker": main_character_ticker,
                    "main_character_icon_url": main_character_icon_url,
                    "standing": p.standing,
                    "labels": labels,
                }
            )

        return characters_data

    my_characters_data = cache_view_pilots_json.get_or_set(generate_characters_data)
    return JsonResponse(my_characters_data, safe=False)


@login_required
@permission_required("standingsrequests.download")
def download_pilot_standings(request):
    logger.info("download_pilot_standings called by %s", request.user)
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = 'attachment; filename="standings.csv"'
    writer = UnicodeWriter(response)
    try:
        contacts = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = ContactSet()

    writer.writerow(
        [
            "character_id",
            "character_name",
            "corporation_id",
            "corporation_name",
            "corporation_ticker",
            "alliance_id",
            "alliance_name",
            "has_scopes",
            "state",
            "main_character_name",
            "main_character_ticker",
            "standing",
            "labels",
        ]
    )

    # lets request make sure all info is there in bulk
    character_contacts = contacts.charactercontact_set.all().order_by("name")
    EveEntity.objects.get_names([p.contact_id for p in character_contacts])

    for pilot_standing in character_contacts:
        char = EveCharacter.objects.get_character_by_id(pilot_standing.contact_id)
        main = ""
        state = ""
        try:
            ownership = CharacterOwnership.objects.get(character=char)
            state = ownership.user.profile.state.name
            main = ownership.user.profile.main_character
            if main is None:
                main_character_name = ""
            else:
                main_character_name = main.character_name
        except CharacterOwnership.DoesNotExist:
            main_character_name = ""
            main = None

        pilot = [
            pilot_standing.contact_id,
            pilot_standing.name,
            char.corporation_id if char else "",
            char.corporation_name if char else "",
            char.corporation_ticker if char else "",
            char.alliance_id if char else "",
            char.alliance_name if char else "",
            StandingRequest.has_required_scopes_for_request(char),
            state,
            main_character_name,
            main.corporation_ticker if main else "",
            pilot_standing.standing,
            ", ".join([label.name for label in pilot_standing.labels.all()]),
        ]

        writer.writerow([str(v) if v is not None else "" for v in pilot])
    return response


@login_required
@permission_required("standingsrequests.view")
def view_groups_standings(request):
    logger.debug("view_group_standings called by %s", request.user)
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contact_set = None
    finally:
        organization = ContactSet.standings_source_entity()
        last_update = contact_set.date if contact_set else None

    if contact_set:
        groups_count = (
            contact_set.corporationcontact_set.count()
            + contact_set.alliancecontact_set.count()
        )
    else:
        groups_count = None

    return render(
        request,
        "standingsrequests/view_groups.html",
        {
            "lastUpdate": last_update,
            "app_title": __title__,
            "operation_mode": SR_OPERATION_MODE,
            "organization": organization,
            "groups_count": groups_count,
        },
    )


@login_required
@permission_required("standingsrequests.view")
def view_groups_standings_json(request):
    logger.debug("view_pilot_standings_json called by %s", request.user)

    def generate_groups_data():
        try:
            contacts = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            contacts = ContactSet()

        corporations_qs = contacts.corporationcontact_set.all().order_by("name")
        eve_corporations = {
            corporation.corporation_id: corporation
            for corporation in EveCorporation.get_many_by_id(
                corporations_qs.values_list("contact_id", flat=True)
            )
        }
        corporations_data = list()
        standing_requests_qs = StandingRequest.objects.filter(
            contact_type_id=CorporationContact.get_contact_type_id()
        )
        for contact in corporations_qs:
            if contact.contact_id in eve_corporations:
                corporation = eve_corporations[contact.contact_id]
                try:
                    standing_request = standing_requests_qs.get(
                        contact_id=contact.contact_id,
                    )
                except StandingRequest.DoesNotExist:
                    main = None
                    has_required_scopes = None
                    state_name = ""
                else:
                    user = standing_request.user
                    main = user.profile.main_character
                    state_name = user.profile.state.name
                    has_required_scopes = (
                        not corporation.is_npc
                        and corporation.user_has_all_member_tokens(
                            user=user, quick_check=True
                        )
                    )
                finally:
                    main_character_name = main.character_name if main else ""
                    main_character_ticker = main.corporation_ticker if main else ""
                    main_character_icon_url = (
                        main.portrait_url(DEFAULT_ICON_SIZE) if main else ""
                    )
                    labels = [label.name for label in contact.labels.all()]

                corporations_data.append(
                    {
                        "corporation_id": corporation.corporation_id,
                        "corporation_name": corporation.corporation_name,
                        "corporation_icon_url": corporation.logo_url(DEFAULT_ICON_SIZE),
                        "alliance_id": corporation.alliance_id,
                        "alliance_name": corporation.alliance_name,
                        "standing": contact.standing,
                        "labels": labels,
                        "has_required_scopes": has_required_scopes,
                        "state": state_name,
                        "main_character_name": main_character_name,
                        "main_character_ticker": main_character_ticker,
                        "main_character_icon_url": main_character_icon_url,
                    }
                )

        alliances_data = list()
        for contact in contacts.alliancecontact_set.all().order_by("name"):
            alliances_data.append(
                {
                    "alliance_id": contact.contact_id,
                    "alliance_name": contact.name,
                    "alliance_icon_url": EveAllianceInfo.generic_logo_url(
                        contact.contact_id, DEFAULT_ICON_SIZE
                    ),
                    "standing": contact.standing,
                    "labels": [label.name for label in contact.labels.all()],
                }
            )

        return {"corps": corporations_data, "alliances": alliances_data}

    my_groups_data = cache_view_groups_json.get_or_set(generate_groups_data)
    return JsonResponse(my_groups_data, safe=False)


###################
# Manage requests #
###################


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_standings(request):
    logger.debug("manage_standings called by %s", request.user)
    context = {
        "app_title": __title__,
        "operation_mode": SR_OPERATION_MODE,
        "organization": ContactSet.standings_source_entity(),
        "requests_count": _standing_requests_to_manage().count(),
        "revocations_count": _revocations_to_manage().count(),
    }
    return render(request, "standingsrequests/manage.html", context,)


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_get_requests_json(request):
    logger.debug("manage_get_requests_json called by %s", request.user)
    requests_qs = _standing_requests_to_manage()
    requests_data = _compose_standing_requests_data(requests_qs)
    return JsonResponse(requests_data, safe=False)


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_get_revocations_json(request):
    logger.debug("manage_get_revocations_json called by %s", request.user)
    revocations_qs = _revocations_to_manage()
    requests_data = _compose_standing_requests_data(revocations_qs)
    return JsonResponse(requests_data, safe=False)


def _standing_requests_to_manage() -> models.QuerySet:
    return (
        StandingRequest.objects.filter(action_date__isnull=True, is_effective=False)
        .select_related("user__profile")
        .order_by("request_date")
    )


def _revocations_to_manage() -> models.QuerySet:
    return (
        StandingRevocation.objects.filter(action_date__isnull=True, is_effective=False)
        .select_related("user__profile")
        .order_by("request_date")
    )


def _compose_standing_requests_data(
    requests_qs: models.QuerySet, quick_check: bool = False
) -> list:
    """composes list of standings requests or revocations based on queryset 
    and returns it
    """
    requests_data = list()

    # preload characters in bulk
    character_ids = requests_qs.exclude(
        contact_type_id=CorporationContact.get_contact_type_id()
    ).values_list("contact_id", flat=True)
    if character_ids:
        eve_characters = {
            character.character_id: character
            for character in EveCharacter.objects.filter(character_id__in=character_ids)
        }
    else:
        eve_characters = list()
    # preload corporations in bulk
    corporation_ids = requests_qs.filter(
        contact_type_id=CorporationContact.get_contact_type_id()
    ).values_list("contact_id", flat=True)
    if corporation_ids:
        eve_corporations = {
            corporation.corporation_id: corporation
            for corporation in EveCorporation.get_many_by_id(corporation_ids)
        }
    else:
        eve_corporations = list()

    for r in requests_qs:
        main_character_name = ""
        main_character_ticker = ""
        main_character_icon_url = ""
        if r.user:
            state_name = r.user.profile.state.name
            main = r.user.profile.main_character
            if main:
                main_character_name = main.character_name
                main_character_ticker = main.corporation_ticker
                main_character_icon_url = main.portrait_url(DEFAULT_ICON_SIZE)
        else:
            state_name = "(no user)"

        if r.is_character:
            if r.contact_id in eve_characters:
                character = eve_characters[r.contact_id]
            else:
                character = EveCharacterHelper(r.contact_id)

            contact_name = character.character_name
            contact_icon_url = character.portrait_url(DEFAULT_ICON_SIZE)
            corporation_id = character.corporation_id
            corporation_name = (
                character.corporation_name if character.corporation_name else ""
            )
            corporation_ticker = (
                character.corporation_ticker if character.corporation_ticker else ""
            )
            alliance_id = character.alliance_id
            alliance_name = character.alliance_name if character.alliance_name else ""
            has_scopes = StandingRequest.has_required_scopes_for_request(
                character=character, user=r.user, quick_check=quick_check
            )

        elif r.is_corporation and r.contact_id in eve_corporations:
            corporation = eve_corporations[r.contact_id]
            contact_icon_url = corporation.logo_url(DEFAULT_ICON_SIZE)
            contact_name = corporation.corporation_name
            corporation_id = corporation.corporation_id
            corporation_name = corporation.corporation_name
            corporation_ticker = corporation.ticker
            alliance_id = None
            alliance_name = ""
            has_scopes = (
                not corporation.is_npc
                and corporation.user_has_all_member_tokens(
                    user=r.user, quick_check=quick_check
                )
            )

        else:
            contact_name = ""
            contact_icon_url = ""
            corporation_id = None
            corporation_name = ""
            corporation_ticker = ""
            alliance_id = None
            alliance_name = ""
            has_scopes = False

        requests_data.append(
            {
                "contact_id": r.contact_id,
                "contact_name": contact_name,
                "contact_icon_url": contact_icon_url,
                "corporation_id": corporation_id,
                "corporation_name": corporation_name,
                "corporation_ticker": corporation_ticker,
                "alliance_id": alliance_id,
                "alliance_name": alliance_name,
                "request_date": r.request_date.isoformat(),
                "action_date": r.action_date.isoformat() if r.action_date else None,
                "has_scopes": has_scopes,
                "state": state_name,
                "main_character_name": main_character_name,
                "main_character_ticker": main_character_ticker,
                "main_character_icon_url": main_character_icon_url,
                "actioned": r.is_actioned,
                "is_effective": r.is_effective,
                "is_corporation": r.is_corporation,
                "is_character": r.is_character,
                "action_by": r.action_by.username if r.action_by else "(System)",
            }
        )

    return requests_data


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_requests_write(request, contact_id):
    contact_id = int(contact_id)
    logger.debug("manage_requests_write called by %s", request.user)
    if request.method == "PUT":
        actioned = 0
        for r in StandingRequest.objects.filter(contact_id=contact_id):
            r.mark_actioned(request.user)
            actioned += 1
        if actioned > 0:
            return HttpResponseNoContent()
        else:
            return Http404()
    elif request.method == "DELETE":
        try:
            standing_request = StandingRequest.objects.get(contact_id=contact_id)
        except StandingRequest.DoesNotExist:
            return Http404()
        else:
            StandingRequest.objects.remove_requests(contact_id)
            if SR_NOTIFICATIONS_ENABLED:
                entity_name = EveEntity.objects.get_name(contact_id)
                title = _("Standing request for %s rejected" % entity_name)
                message = _(
                    "Your standing request for '%s' "
                    "has been rejected by %s." % (entity_name, request.user)
                )
                notify(user=standing_request.user, title=title, message=message)

            return HttpResponseNoContent()
    else:
        return Http404()


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_revocations_write(request, contact_id):
    contact_id = int(contact_id)
    logger.debug(
        "manage_revocations_write called by %s for contact_id %d",
        request.user,
        contact_id,
    )
    if request.method == "PUT":
        actioned = 0
        for r in StandingRevocation.objects.filter(contact_id=contact_id):
            r.mark_actioned(request.user)
            actioned += 1
        if actioned > 0:
            return HttpResponseNoContent()
        else:
            return Http404
    elif request.method == "DELETE":
        try:
            standing_revocation = StandingRevocation.objects.get(contact_id=contact_id)
        except StandingRevocation.DoesNotExist:
            return Http404()
        else:
            StandingRevocation.objects.filter(contact_id=contact_id).delete()
            if SR_NOTIFICATIONS_ENABLED and standing_revocation.user:
                entity_name = EveEntity.objects.get_name(contact_id)
                title = _("Standing revocation for %s rejected" % entity_name)
                message = _(
                    "Your standing revocation for '%s' "
                    "has been rejected by %s." % (entity_name, request.user)
                )
                notify(user=standing_revocation.user, title=title, message=message)

            return HttpResponseNoContent()

    else:
        return Http404()


###################
# View requests #
###################


@login_required
@permission_required("standingsrequests.affect_standings")
def view_active_requests(request):
    context = {
        "app_title": __title__,
        "operation_mode": SR_OPERATION_MODE,
        "organization": ContactSet.standings_source_entity(),
        "requests_count": _standing_requests_to_view().count(),
    }
    return render(request, "standingsrequests/requests.html", context)


@login_required
@permission_required("standingsrequests.affect_standings")
def view_requests_json(request):

    response_data = _compose_standing_requests_data(
        _standing_requests_to_view(), quick_check=True
    )
    return JsonResponse(response_data, safe=False)


def _standing_requests_to_view() -> models.QuerySet:
    return (
        StandingRequest.objects.filter(is_effective=True)
        .select_related("user__profile")
        .order_by("-request_date")
    )


@login_required
@permission_required("standingsrequests.affect_standings")
@token_required(new=False, scopes=ContactSet.required_esi_scope())
def view_auth_page(request, token):
    source_entity = ContactSet.standings_source_entity()
    char_name = EveEntity.objects.get_name(STANDINGS_API_CHARID)
    if not source_entity:
        messages_plus.error(
            request,
            format_html(
                _(
                    "The configured character <strong>%s</strong> does not belong "
                    "to an alliance and can therefore not be used "
                    "to setup alliance standings. "
                    "Please configure a character that has an alliance."
                )
                % char_name,
            ),
        )
    elif token.character_id == STANDINGS_API_CHARID:
        update_all.delay(user_pk=request.user.pk)
        messages_plus.success(
            request,
            format_html(
                _(
                    "Token for character <strong>%s</strong> has been setup "
                    "successfully and the app has started pulling standings "
                    "from <strong>%s</strong>."
                )
                % (char_name, source_entity.name),
            ),
        )
    else:
        messages_plus.error(
            request,
            _(
                "Failed to setup token for configured character "
                "%(char_name)s (id:%(standings_api_char_id)s). "
                "Instead got token for different character: "
                "%(token_char_name)s (id:%(token_char_id)s)"
            )
            % {
                "char_name": char_name,
                "standings_api_char_id": STANDINGS_API_CHARID,
                "token_char_name": EveEntity.objects.get_name(token.character_id),
                "token_char_id": token.character_id,
            },
        )
    return redirect("standingsrequests:index")


@login_required
@permission_required(StandingRequest.REQUEST_PERMISSION_NAME)
@token_required_by_state(new=False)
def view_requester_add_scopes(request, token):
    messages_plus.success(
        request,
        _("Successfully added token with required scopes for %(char_name)s")
        % {"char_name": EveEntity.objects.get_name(token.character_id)},
    )
    return redirect("standingsrequests:index")



from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

from .. import __title__
from ..utils import LoggerAddTag

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EveEntityHelper:
    @staticmethod
    def get_owner_from_character_id(character_id):
        """
        Attempt to get the character owner from the given character_id
        :param character_id: int character ID to get the owner for
        :return: User (django) or None
        """
        char = EveCharacter.objects.get_character_by_id(character_id)
        if char is not None:
            try:
                ownership = CharacterOwnership.objects.get(character=char)
                return ownership.user

            except CharacterOwnership.DoesNotExist:
                return None
        else:
            return None

    @staticmethod
    def get_characters_by_user(user):
        return EveCharacter.objects.filter(
            character_ownership__user=user
        ).select_related("character_ownership__user")

    @staticmethod
    def is_character_owned_by_user(character_id, user):
        try:
            CharacterOwnership.objects.get(
                user=user, character__character_id=character_id
            )

            return True
        except CharacterOwnership.DoesNotExist:
            return False

    @staticmethod
    def get_state_of_character(char):
        try:
            ownership = CharacterOwnership.objects.get(
                character__character_id=char.character_id
            )
            return ownership.user.profile.state.name

        except CharacterOwnership.DoesNotExist:
            return None



import csv
import codecs
from io import StringIO


class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    Credit: https://docs.python.org/2/library/csv.html#examples
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8") for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



from concurrent.futures import ThreadPoolExecutor

from bravado.exception import HTTPError

from django.contrib.auth.models import User
from django.core.cache import cache

from allianceauth.eveonline.models import EveCharacter
from allianceauth.eveonline.evelinks import eveimageserver
from allianceauth.services.hooks import get_extension_logger

from .. import __title__
from .esi_fetch import esi_fetch, _esi_client
from ..utils import LoggerAddTag


logger = LoggerAddTag(get_extension_logger(__name__), __title__)

MAX_WORKERS = 10


class EveCorporation:
    CACHE_PREFIX = "STANDINGS_REQUESTS_EVECORPORATION_"
    CACHE_TIME = 60 * 60  # 60 minutes

    def __init__(self, **kwargs):
        self.corporation_id = int(kwargs.get("corporation_id"))
        self.corporation_name = kwargs.get("corporation_name")
        self.ticker = kwargs.get("ticker")
        self.member_count = kwargs.get("member_count")
        self.ceo_id = kwargs.get("ceo_id")
        self.alliance_id = kwargs.get("alliance_id")
        self.alliance_name = kwargs.get("alliance_name")

    def __str__(self):
        return self.corporation_name

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, type(self))
            and self.corporation_id == o.corporation_id
            and self.corporation_name == o.corporation_name
            and self.ticker == o.ticker
            and self.member_count == o.member_count
            and self.ceo_id == o.ceo_id
            and self.alliance_id == o.alliance_id
            and self.alliance_name == o.alliance_name
        )

    @property
    def is_npc(self) -> bool:
        """returns true if this corporation is an NPC, else false"""
        return self.corporation_is_npc(self.corporation_id)

    @staticmethod
    def corporation_is_npc(corporation_id: int) -> bool:
        """returns true if this corporation is an NPC, else false"""
        return 1000000 <= corporation_id <= 2000000

    def logo_url(self, size: int = eveimageserver._DEFAULT_IMAGE_SIZE) -> str:
        return eveimageserver.corporation_logo_url(self.corporation_id, size)

    def member_tokens_count_for_user(
        self, user: User, quick_check: bool = False
    ) -> int:
        """returns the number of character tokens the given user owns 
        for this corporation

        Params:
        - user: user owning the characters
        - quick: if True will not check if tokens are valid to save time
        """
        from ..models import StandingRequest

        return sum(
            [
                1
                for character in EveCharacter.objects.filter(
                    character_ownership__user=user
                )
                .select_related("character_ownership__user__profile__state")
                .filter(corporation_id=self.corporation_id)
                if StandingRequest.has_required_scopes_for_request(
                    character=character, user=user, quick_check=quick_check
                )
            ]
        )

    def user_has_all_member_tokens(self, user: User, quick_check: bool = False) -> bool:
        """returns True if given user owns same amount of token than there are 
        member characters in this corporation, else False

        Params:
        - user: user owning the characters
        - quick: if True will not check if tokens are valid to save time
        """
        return (
            self.member_count is not None
            and self.member_tokens_count_for_user(user=user, quick_check=quick_check)
            >= self.member_count
        )

    @classmethod
    def get_by_id(cls, corporation_id: int, ignore_cache: bool = False) -> object:
        """Get a corporation from the cache or ESI if not cached
        Corps are cached for 3 hours
        
        Params
        - corporation_id: int corporation ID to get
        - ignore_cache: when true will always get fresh from API
        
        Returns corporation object or None
        """
        logger.debug("Getting corporation by id %d", corporation_id)
        my_cache_key = cls._get_cache_key(corporation_id)
        corporation = cache.get(my_cache_key)
        if corporation is None or ignore_cache:
            logger.debug("Corp not in cache or ignoring cache, fetching")
            corporation = cls.fetch_corporation_from_api(corporation_id)
            if corporation is not None:
                cache.set(my_cache_key, corporation, cls.CACHE_TIME)
        else:
            logger.debug("Retreving corporation %s from cache", corporation_id)
        return corporation

    @classmethod
    def _get_cache_key(cls, corporation_id: int) -> str:
        return cls.CACHE_PREFIX + str(corporation_id)

    @classmethod
    def fetch_corporation_from_api(cls, corporation_id):
        from ..models import EveEntity

        logger.debug(
            "Attempting to fetch corporation from ESI with id %s", corporation_id
        )
        try:
            info = esi_fetch(
                "Corporation.get_corporations_corporation_id",
                args={"corporation_id": corporation_id},
            )
        except HTTPError:
            logger.exception(
                "Failed to fetch corporation from ESI with id %i", corporation_id
            )
            return None

        else:
            args = {
                "corporation_id": corporation_id,
                "corporation_name": info["name"],
                "ticker": info["ticker"],
                "member_count": info["member_count"],
                "ceo_id": info["ceo_id"],
            }
            if "alliance_id" in info and info["alliance_id"]:
                args["alliance_id"] = info["alliance_id"]
                args["alliance_name"] = EveEntity.objects.get_name(info["alliance_id"])

            return cls(**args)

    @classmethod
    def thread_fetch_corporation(cls, corporation_id: int) -> object:
        """Gets one corporation by ID and returns it - used for threads"""
        return EveCorporation.get_by_id(corporation_id)

    @classmethod
    def get_many_by_id(cls, corporation_ids: list) -> list:
        """Returns multiple corporations by ID
        
        Fetches requested corporations from cache or API as needed. 
        Uses threads to fetch them in parallel.
        """
        if len(corporation_ids) == 0:
            return []
        else:
            _esi_client()  # make sure client is loaded before starting threads
            logger.info(
                "Starting to fetch the %d corporations from ESI with up to %d workers",
                len(corporation_ids),
                MAX_WORKERS,
            )
            with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
                futures = [
                    executor.submit(cls.thread_fetch_corporation, corporation_id)
                    for corporation_id in corporation_ids
                ]
                logger.info(
                    "Waiting for all threads fetching corporations to complete..."
                )

            logger.info(
                "Completed fetching %d corporations from ESI", len(corporation_ids)
            )
            return [f.result() for f in futures]



from django.core.management.base import BaseCommand

from ...models import ContactSet


def get_input(text):
    """wrapped input to enable unit testing / patching"""
    return input(text)


class Command(BaseCommand):
    help = "Create standing request for alts with standing in game"

    def _create_requests(self):
        created_counter = (
            ContactSet.objects.latest().generate_standing_requests_for_blue_alts()
        )
        self.stdout.write(f"Created a total of {created_counter} standing requests.")

    def handle(self, *args, **options):
        self.stdout.write(
            "This command will automatically create accepted standings requests for "
            "alt characters on Auth that already have blue standing in-game."
        )
        user_input = get_input("Are you sure you want to proceed? (Y/n)?")
        if user_input == "Y":
            self.stdout.write("Starting update. Please stand by.")
            self._create_requests()
            self.stdout.write(self.style.SUCCESS("Process completed!"))
        else:
            self.stdout.write(self.style.WARNING("Aborted"))
